
// References
//
// [Radzik1992], "Newton's Method for Fractional Combinatorial Optimization"
//     Tomasz Radzik, 1992.

#include <algorithm>
#include <numeric>

#include "GraphClusteringLFCO.h"

namespace Clustering {

GraphClusteringLFCO::GraphClusteringLFCO() {
}

void GraphClusteringLFCO::Initialize(
	const gmm::row_matrix<gmm::wsvector<double> >& Kcost,
	const gmm::row_matrix<gmm::wsvector<double> >& Kweight) {
	GraphClustering::Initialize(Kweight);

	obj_a.resize(edgemap.size());
	obj_b.resize(edgemap.size());
	for (unsigned int ei = 0; ei < edgemap.size(); ++ei) {
		obj_a[ei] = Kcost(edgemap[ei].first, edgemap[ei].second);
		obj_b[ei] = Kweight(edgemap[ei].first, edgemap[ei].second);
	}
}

void GraphClusteringLFCO::Initialize(const gmm::dense_matrix<double>& Kcost,
	const gmm::dense_matrix<double>& Kweight) {
	GraphClustering::Initialize(Kweight);

	obj_a.resize(edgemap.size());
	obj_b.resize(edgemap.size());
	for (unsigned int ei = 0; ei < edgemap.size(); ++ei) {
		obj_a[ei] = Kcost(edgemap[ei].first, edgemap[ei].second);
		obj_b[ei] = Kweight(edgemap[ei].first, edgemap[ei].second);
	}
}

void GraphClusteringLFCO::Solve() {
	std::vector<double> deltas;
	std::vector<double> h_deltas;
	double delta = 0;
	double* obj_arr = new double[obj.size()];
	std::vector<double> sol_last;
	for (int iter = 1; true; ++iter) {
		std::cout << "[lfco] iter " << iter << std::endl;

		// Set the objective to the current shift minus a small number
		for (unsigned int ei = 0; ei < obj_a.size(); ++ei)
			obj_arr[ei] = obj_a[ei] - delta * obj_b[ei];
		std::copy(obj_arr, obj_arr + obj_a.size(), obj.begin());
		si->setObjective(obj_arr);

		// First iteration if theta_right is infinity
		std::cout << "FIXED with delta = " << delta << std::endl;
		unsigned int iters = SolveFixed();
		std::cout << "      FIXED obj " << Objective()
			<< ", took " << (iters == 0 ? max_iter : iters)
			<< " iterations" << std::endl;

		// Convergence condition: found a root
		double h_delta = GraphClustering::Objective();
		h_deltas.push_back(h_delta);
		std::cout << "deltas: " << gmm::row_vector(deltas) << std::endl;
		std::cout << "h(delta): " << gmm::row_vector(h_deltas) << std::endl;
		if (fabs(h_delta) < 1e-5)
			break;

		// Newton update on delta
		const std::vector<double>& sol = Solution();
		sol_last = sol;
		delta = std::inner_product(obj_a.begin(), obj_a.end(), sol.begin(), 0.0) /
			std::inner_product(obj_b.begin(), obj_b.end(), sol.begin(), 0.0);

		deltas.push_back(delta);
		std::cout << "deltas: " << gmm::row_vector(deltas) << std::endl;
	}
	edgevar = sol_last;
}

double GraphClusteringLFCO::Objective() const {
	double objval_num = 0.0;
	double objval_denum = 0.0;
	for (unsigned int ei = 0; ei < obj_a.size(); ++ei) {
		objval_num += edgevar[ei] * obj_a[ei];
		objval_denum += edgevar[ei] * obj_b[ei];
	}
	return (objval_num / objval_denum);
}

const std::vector<double>& GraphClusteringLFCO::Solution() const {
	return (GraphClustering::Solution());
}

bool GraphClusteringLFCO::IsFractionalSolution() const {
	return (GraphClustering::IsFractionalSolution());
}

const std::vector<std::pair<unsigned int, unsigned int> >&
GraphClusteringLFCO::Edgemap() const {
	return (GraphClustering::Edgemap());
}

}
