/* mexgclustering.cpp - LP Graph Clustering
 *
 * Author: Sebastian Nowozin <sebastian.nowozin@tuebingen.mpg.de>
 * Date: 10th June 2008
 *
 * Usage:
 *
 * [Z, part, obj, is_integral, theta_range, E_stab] = mexgclustering(K, options);
 *   for normal linear objectives or
 * [Z, part, obj, is_integral] = mexgclustering({Kcost, Kweight});
 *   for fractional objectives.  We must have Kweight >= 0, Kcost is
 *   arbitrary.
 *
 ***
 * Input
 *    K: (n,n) Centered sample similarity matrix (eg. mean(K(:)) \approx 0).
 *       Must be symmetric.  If dense, a completely connected graph will be
 *       created, even for zero elements.  If sparse, edges are only created
 *       for the given values.
 *       (or)
 *       (1,2) cellarray of two (n,n) matrices Kcost, Kweight.
 *       We must have: Kcost(i,j) ~= 0 ==> Kweight(i,j) ~= 0.
 *       The objective minimized is the fraction (Kcost'Z) / (Kweight'Z).
 *       Note that stability analysis and theta_ranges are not yet available
 *       for fractional objectives.
 *    options: (optional) structure with the following elements
 *       .heuristic_kl: (optional), (1,1) either 0 (use LP) or 1 (use
 *          Kernighan-Lin heuristic).
 *       .do_separate_oddwheel: (optional), (1,1) either 0 (no use of oddwheel
 *          inequalities), or 1 (use oddwheel inequalities).
 *       .theta_range: (optional), (1,2) range [a;b] with a <= b, and
 *          a,b \in (\R \cup \{-\infty,+\infty\}).  The solution path will be
 *          traced for all values in theta.  If a < b, Z will be a cellarray
 *          of structures, see below.
 *       .max_iter: (optional), (1,1) integer, if <= 0 unlimited outer
 *          relaxed-LP iterations are allowed, if > 0, its limited to at most
 *          these.
 *       .Dsens: (optional), (s,|E|) matrix of s direction vectors along the
 *          edge weights.  |E| should have (n(n-1))/2 elements if K is dense,
 *          otherwise it should be nnz(squareform(sparse(K-diag(diag(K))))).
 *          Once an optimal solution is obtained, for each of the s
 *          directions, a stability range is calculated an put into E_stab.
 *
 * Output
 *    Z: If no options or options.theta_range is given or if
 *       theta_range(1)=theta_range(2), then: (n,n) matrix with elements
 *       0 or 1.  Z_{i,j} = 0 means sample i and j are in the same cluster,
 *       Z_{i,j} = 1 that they are in different clusters.  Z is symmetric and
 *       dense.  NOTE: All elements in Z(i,j) for which no edge exists in the
 *       graph are assigned a NaN (not a number) value.  p = 1 for the
 *       remaining output parameters.
 *
 *       If a theta_range with theta_range(1) < theta_range(2) is given, then
 *       Z is a cellarray with length p, each element has a (n,n) matrix with
 *       the above interpretation
 *    part: (optional) (n,p) partition vector(s), assigning each sample a
 *       discrete partition index.
 *    obj: (optional) (1,p) scalar, achieved objective(s).
 *       FIXME: for now, ignore the objective, its all NaN
 *    is_integral: (optional) (1,p) flag(s), >=1 if respective solution is
 *       optimal, 0 if it is only a fractional relaxation and obj is a lower
 *       bound.  If >= 1, it is the number of connected components in the
 *       clustering.
 *    theta_range: (optional) (p,2) stability range, that is the p'th Z is the
 *       optimal solution for any linear shift K+theta*spones(size(K)), where
 *       theta >= theta_range(1) and theta <= theta_range(2).
 *    E_stab: (optional) (n,2) per edge stability ranges with values in R \cup
 *       \{-\infty,\infty\}.  We guarantee that if a single edge weight is
 *       perturbed within the respective range, the solution remains optimal.
 *       This is only available for a single solution output.
 *       If options.Dsens is used, E_stab is (s,2) and contains the global
 *       stability ranges for each direction vector.
 *
 */

#include <vector>
#include <algorithm>
#include <limits>
#include <assert.h>
#include <mex.h>
#include <gmm/gmm.h>

#include "GraphClustering.h"
#include "GraphClusteringLFCO.h"

static mxArray* CreateZfromSolution(const std::vector<double>& solution,
	const std::vector<std::pair<unsigned int, unsigned int> >& edgemap,
	unsigned int Krows) {
	unsigned int Kcols = Krows;
	mxArray* Z = mxCreateNumericMatrix(Krows, Kcols, mxDOUBLE_CLASS, mxREAL);
	double* Zout = (double*) mxGetPr(Z);
	unsigned int ke = Krows * Kcols;
	for (unsigned int n = 0; n < ke; ++n)
		Zout[n] = std::numeric_limits<double>::signaling_NaN();

	for (unsigned int n = 0; n < solution.size(); ++n) {
		unsigned int r = edgemap[n].first;
		unsigned int c = edgemap[n].second;
		Zout[c*Krows + r] = solution[n];
		Zout[r*Krows + c] = solution[n];
	}
	for (unsigned int n = 0; n < Krows; ++n)
		Zout[n*Krows + n] = 0;	// Diagonal is zero.

	return (Z);
}

static unsigned int LabelPartitions(double* partout, unsigned int N,
	const std::vector<double>& solution,
	const std::vector<std::pair<unsigned int, unsigned int> >& edgemap) {
	unsigned int components = 0;
	std::fill(partout, partout + N, -1.0);
	while (1) {
		// 1. Search for first unlabeled node
		unsigned int node;
		for (node = 0; node < N; ++node) {
			if (partout[node] < 0.0)
				break;
		}
		if (node == N)
			break;	// no unlabeled nodes left

		// Node is unlabeled, now assign it a unique component
		partout[node] = components + 1;
		components += 1;

		// Update the adjacent nodes recursively until no updates are left
		bool updates_done = true;
		while (updates_done) {
			updates_done = false;
			for (unsigned int n = 0; n < solution.size(); ++n) {
				unsigned int n1 = edgemap[n].first;
				unsigned int n2 = edgemap[n].second;
				if (partout[n1] < 0.0 && partout[n2] < 0.0)
					continue;

				// XXX: Treat nodes with strictly positive edge variables as
				// if they are set to one.
				if (fabs(solution[n]) > 1e-5)
					continue;

				if (partout[n1] > 0.0 && partout[n2] < 0.0) {
					partout[n2] = partout[n1];
					updates_done = true;
				} else if (partout[n1] < 0.0 && partout[n2] > 0.0) {
					partout[n1] = partout[n2];
					updates_done = true;
				}
			}
		}
	}
	return (components);
}

static void ConvertToGMM(const mxArray* mx,
	gmm::row_matrix<gmm::wsvector<double> >& K) {
	unsigned int Krows = mxGetM(mx);
	unsigned int Kcols = mxGetN(mx);
	K.resize(Krows, Kcols);
	if (mxIsSparse(mx)) {
		/* Sparse matrix: sparse graph
		 *
		 * (from Matlab documentation)
		 * "Sparse matrices have a different storage convention from that of
		 * full matrices in MATLAB. The parameters pr and pi are still arrays
		 * of double-precision numbers, but these arrays contain only nonzero
		 * data elements. There are three additional parameters: nzmax, ir,
		 * and jc.
		 *
		 * * nzmax is an integer that contains the length of ir, pr, and, if
		 * it exists, pi. It is the maximum possible number of nonzero
		 * elements in the sparse matrix.
		 *
		 * * ir points to an integer array of length nzmax containing the row
		 * indices of the corresponding elements in pr and pi.
		 *
		 * * jc points to an integer array of length n+1, where n is the
		 * number of columns in the sparse matrix. The jc array contains
		 * column index information. If the j'th column of the sparse matrix
		 * has any nonzero elements, jc[j] is the index in ir and pr (and pi
		 * if it exists) of the first nonzero element in the jth column, and
		 * jc[j+1] - 1 is the index of the last nonzero element in that
		 * column. For the jth column of the sparse matrix, jc[j] is the total
		 * number of nonzero elements in all preceding columns. The last
		 * element of the jc array, jc[n], is equal to nnz, the number of
		 * nonzero elements in the entire sparse matrix. If nnz is less than
		 * nzmax, more nonzero entries can be inserted into the array without
		 * allocating additional storage."
		 */
		const double* kP = (const double*) mxGetPr(mx);
		unsigned int nzmax = (unsigned int) mxGetNzmax(mx);
		mwIndex* ir = mxGetIr(mx);
		mwIndex* jc = mxGetJc(mx);
		unsigned int last_column_nnz = 0;
		for (unsigned int j = 0; j < Kcols; ++j) {
			// Special case: empty column
			if (jc[j] == jc[j + 1])
				continue;

			unsigned int col_end = jc[j + 1] - 1;
			for (unsigned int ci = jc[j]; ci <= col_end; ++ci) {
				assert(ir[ci] >= 0 && ir[ci] < Krows);
				assert(j >= 0 && j < Kcols);
				K(ir[ci], j) = kP[ci];
			}
		}
	}
}

static void ConvertToGMM(const mxArray* mx,
	gmm::dense_matrix<double>& Kdense) {
	unsigned int Krows = mxGetM(mx);
	unsigned int Kcols = mxGetN(mx);

	/* Dense matrix: completely connected graph
	 */
	Kdense.resize(Krows, Kcols);
	const double* kP = (const double*) mxGetPr(mx);
	for (unsigned int r = 0; r < Krows; ++r)
		for (unsigned int c = 0; c < Kcols; ++c)
			Kdense(r,c) = kP[c*Krows+r];	// column-wise
}

/* [Z, part, obj, is_integral, theta_range, E_stab] = mexgclustering(K, options);
 */
void mexFunction(int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[]) {
	if (nlhs < 1 || nlhs > 6) {
		mexErrMsgTxt("Wrong number of left hand side parameters (must be 1 to 6).\n");
		return;
	}
	if (nrhs < 1 || nrhs > 2) {
		mexErrMsgTxt("Wrong number of right hand side parameters (must be 1 or 2).\n");
		return;
	}

	/* Obtain K
	 */
	if (mxIsCell(prhs[0])) {
		if (nlhs < 1 || nlhs > 4) {
			mexErrMsgTxt("Fractional objectives only support one to four "
				"left hand side parameters.\n");
			return;
		}
		if (nrhs > 1) {
			mexErrMsgTxt("Fractional objectives only support one right hand "
				"side parameter.\n");
			return;
		}
		if (mxGetM(prhs[0]) != 1 || mxGetN(prhs[0]) != 2) {
			mexErrMsgTxt("If K is a cellarray, it must be a (1,2) cellarray.\n");
			return;
		}

		// Main clustering object
		Clustering::GraphClusteringLFCO gc;
		const mxArray** karr = (const mxArray**) mxGetPr(prhs[0]);
		unsigned int Krows = mxGetM(karr[0]);
		if (mxIsSparse(karr[0])) {
			gmm::row_matrix<gmm::wsvector<double> > Kcost;
			gmm::row_matrix<gmm::wsvector<double> > Kweight;
			ConvertToGMM(karr[0], Kcost);
			ConvertToGMM(karr[1], Kweight);
			gc.Initialize(Kcost, Kweight);
		} else {
			gmm::dense_matrix<double> Kcost;
			gmm::dense_matrix<double> Kweight;
			ConvertToGMM(karr[0], Kcost);
			ConvertToGMM(karr[1], Kweight);
			gc.Initialize(Kcost, Kweight);
		}
		gc.Solve();

		// Write outputs
		const std::vector<std::pair<unsigned int, unsigned int> >& edgemap =
			gc.Edgemap();

		// Z
		if (nlhs >= 1)
			plhs[0] = CreateZfromSolution(gc.Solution(), edgemap, Krows);

		// part
		unsigned int components = 0;
		if (nlhs >= 2) {
			double* partout = NULL;
			plhs[1] = mxCreateNumericMatrix(Krows, 1, mxDOUBLE_CLASS, mxREAL);
			partout = (double*) mxGetPr(plhs[1]);
			std::fill(partout, partout + Krows, -1.0);
			components = LabelPartitions(partout, Krows,
				gc.Solution(), edgemap);
		}

		// obj
		if (nlhs >= 3) {
			plhs[2] = mxCreateNumericMatrix(1, 1, mxDOUBLE_CLASS, mxREAL);
			double* obj_out = (double*) mxGetPr(plhs[2]);
			obj_out[0] = gc.Objective();
		}

		// is_integral
		if (nlhs >= 4) {
			plhs[3] = mxCreateNumericMatrix(1, 1, mxDOUBLE_CLASS, mxREAL);
			double* is_integral_out = (double*) mxGetPr(plhs[3]);
			is_integral_out[0] = gc.IsFractionalSolution() ? 0 : components;
		}

		return;
	} else {
		unsigned int Krows = mxGetM(prhs[0]);
		unsigned int Kcols = mxGetN(prhs[0]);
		if (Krows != Kcols) {
			mexErrMsgTxt("Input matrix K must be square.\n");
			return;
		}

		bool heuristic_kl = false;	// Use KL heuristic
		bool do_separate_oddwheel = true;
		double theta_left = 0.0;
		double theta_right = 0.0;
		bool has_path = false;
		int max_iter = -1;
		bool use_Dsens = false;
		std::vector<std::vector<double> > Dsens;
		if (nrhs >= 2) {
			if (mxIsStruct(prhs[1]) == false) {
				mexErrMsgTxt("options must be a structure.\n");
				return;
			}
			int fn_hkl = mxGetFieldNumber(prhs[1], "heuristic_kl");
			if (fn_hkl >= 0) {
				const mxArray* options_hkl = mxGetFieldByNumber(prhs[1], 0, fn_hkl);
				double* options_hklv = mxGetPr(options_hkl);
				heuristic_kl = (options_hklv[0] > 0.0);
			}

			int fn_dso = mxGetFieldNumber(prhs[1], "do_separate_oddwheel");
			if (fn_dso >= 0) {
				const mxArray* options_st = mxGetFieldByNumber(prhs[1], 0, fn_dso);
				double* options_dsow = mxGetPr(options_st);
				do_separate_oddwheel = (options_dsow[0] > 0.0);
			}

			// Theta range option
			int fn_tr = mxGetFieldNumber(prhs[1], "theta_range");
			if (fn_tr >= 0) {
				const mxArray* options_tr = mxGetFieldByNumber(prhs[1], 0, fn_tr);
				if (options_tr != NULL) {
					if (nlhs >= 6) {
						mexErrMsgTxt("You cannot use options.theta_range and E_stab at "
							"the same time.\n");
						return;
					}
					if (mxGetM(options_tr) != 1 || mxGetN(options_tr) != 2) {
						mexErrMsgTxt("The options.theta_range parameter must be a "
							"(1,2) double vector.\n");
						return;
					}
					double* options_trp = mxGetPr(options_tr);
					theta_left = options_trp[0];
					theta_right = options_trp[1];
				}
				has_path = (theta_left < theta_right);
			}

			// Maximum iteration option
			int fn_mi = mxGetFieldNumber(prhs[1], "max_iter");
			if (fn_mi >= 0) {
				const mxArray* options_mi = mxGetFieldByNumber(prhs[1], 0, fn_mi);
				if (options_mi != NULL) {
					if (mxGetM(options_mi) != 1 || mxGetN(options_mi) != 1) {
						mexErrMsgTxt("The options.max_iter parameter must be a "
							"(1,1) double vector.\n");
						return;
					}
					double* options_mip = mxGetPr(options_mi);
					max_iter = static_cast<int>(options_mip[0]);
				}
			}

			// Direction sensitivity analysis
			int fn_ds = mxGetFieldNumber(prhs[1], "Dsens");
			if (fn_ds >= 0) {
				const mxArray* options_ds = mxGetFieldByNumber(prhs[1], 0, fn_ds);
				if (options_ds != NULL) {
					unsigned int s = mxGetM(options_ds);
					unsigned int edgecount = mxGetN(options_ds);
					double* options_dsp = mxGetPr(options_ds);

					Dsens.resize(s);
					for (unsigned int dsn = 0; dsn < s; ++dsn) {
						Dsens[dsn].resize(edgecount);
						for (unsigned int ei = 0; ei < edgecount; ++ei)
							Dsens[dsn][ei] = options_dsp[dsn + ei*s];
					}
					use_Dsens = true;
				}
			}
		}

		// Main clustering object
		Clustering::GraphClustering gc(do_separate_oddwheel, max_iter);
		if (mxIsSparse(prhs[0])) {
			gmm::row_matrix<gmm::wsvector<double> > K;
			ConvertToGMM(prhs[0], K);
			gc.Initialize(K);
		} else {
			gmm::dense_matrix<double> K;
			ConvertToGMM(prhs[0], K);
			gc.Initialize(K);
		}

		// Do the clustering
		if (heuristic_kl == false) {
			gc.Solve(theta_left, theta_right);
		} else {
			gc.SolveHeuristicallyKL();
		}

		/* Create output matrix/matrices Z/Zs
		 */
		const std::vector<std::pair<unsigned int, unsigned int> >& edgemap =
			gc.Edgemap();
		unsigned int p = gc.ThetaPath().size();
#if 0
		// FIXME: remove the following two lines
		has_path = false;
		p = 1;
#endif

		assert(has_path || p == 1);
		if (nlhs >= 1) {
			if (has_path) {
				const std::vector<std::pair<double, double> >& theta_path = gc.ThetaPath();
				const std::vector<std::vector<double> >& theta_sol = gc.ThetaSolutions();

				mxArray* Zs = plhs[0] = mxCreateCellMatrix(1, theta_path.size());
				assert(Zs != NULL);

				for (unsigned int sol_n = 0; sol_n < theta_path.size(); ++sol_n) {
					mxArray* Z = CreateZfromSolution(theta_sol[sol_n], edgemap, Krows);
					mxSetCell(Zs, sol_n, Z);
				}

			} else {
				plhs[0] = CreateZfromSolution(gc.Solution(), edgemap, Krows);
			}
		}

		/* If we are asked to create a partition map(s), do so
		 */
		if (nlhs >= 2) {
			double* partout = NULL;
			plhs[1] = mxCreateNumericMatrix(Krows, p, mxDOUBLE_CLASS, mxREAL);
			partout = (double*) mxGetPr(plhs[1]);
			std::fill(partout, partout + Krows*p, -1.0);
			if (has_path) {
				for (unsigned int sol_n = 0; sol_n < p; ++sol_n) {
					unsigned int components = LabelPartitions(
						partout + sol_n * Krows, Krows,
						gc.ThetaSolutions()[sol_n], edgemap);
					if (gc.IsFractionalSolution(gc.ThetaSolutions()[sol_n])) {
						std::cout << "Solution for range [" << gc.ThetaPath()[sol_n].first
							<< " ; " << gc.ThetaPath()[sol_n].second << "] is "
							<< "not integral." << std::endl;
					} else {
						std::cout << "Solution for range [" << gc.ThetaPath()[sol_n].first
							<< " ; " << gc.ThetaPath()[sol_n].second << "] has "
							<< components << " connected components."
							<< std::endl;
					}
				}
			} else {
				unsigned int components = LabelPartitions(partout, Krows,
					gc.Solution(), edgemap);
				std::cout << "Found " << components << " connected components."
					<< std::endl;
			}
		}

		/* Array of objective values
		 */
		if (nlhs >= 3) {
			plhs[2] = mxCreateNumericMatrix(1, p, mxDOUBLE_CLASS, mxREAL);
			double* obj_out = (double*) mxGetPr(plhs[2]);
			if (has_path) {
				std::fill(obj_out, obj_out + p,
					std::numeric_limits<double>::signaling_NaN());
			} else {
				obj_out[0] = gc.Objective();
			}
		}

		/* Integrality/component-count flag(s)
		 */
		const std::vector<bool>& iters_exceeded = gc.ThetaItersExceeded();
		if (nlhs >= 4) {
			plhs[3] = mxCreateNumericMatrix(1, p, mxDOUBLE_CLASS, mxREAL);
			double* is_integral_out = (double*) mxGetPr(plhs[3]);
			double* partout = new double[gc.Solution().size()];
			if (has_path) {
				for (unsigned int sol_n = 0; sol_n < p; ++sol_n) {
					unsigned int components = LabelPartitions(
						partout, Krows, gc.ThetaSolutions()[sol_n], edgemap);
					is_integral_out[sol_n] = iters_exceeded[sol_n] ? 0 :
						gc.IsFractionalSolution(gc.ThetaSolutions()[sol_n]) ?
							0 : components;
				}
			} else {
				unsigned int components = LabelPartitions(
					partout, Krows, gc.Solution(), edgemap);
				is_integral_out[0] = iters_exceeded[0] ? 0 :
					(gc.IsFractionalSolution() ? 0 : components);
			}
			delete[] partout;
		}

		/* theta_range
		 */
		if (nlhs >= 5) {
			plhs[4] = mxCreateNumericMatrix(p, 2, mxDOUBLE_CLASS, mxREAL);
			double* theta_range_p = (double*) mxGetPr(plhs[4]);
			for (unsigned int sol_n = 0; sol_n < p; ++sol_n) {
				theta_range_p[0 * p + sol_n] = gc.ThetaPath()[sol_n].first;
				theta_range_p[1 * p + sol_n] = gc.ThetaPath()[sol_n].second;
			}
		}

		/* E_stab
		 */
		if (nlhs >= 6 && use_Dsens) {
			// Use specified directions
			assert(gc.Solution().size() == Dsens[0].size());

			plhs[5] = mxCreateNumericMatrix(Dsens.size(), 2, mxDOUBLE_CLASS, mxREAL);
			double* E_stab_p = (double*) mxGetPr(plhs[5]);
			for (unsigned int s = 0; s < Dsens.size(); ++s) {
				double theta_min;
				double theta_max;
				gc.ComputeStabilityRangeLP(Dsens[s], theta_min, theta_max);
				E_stab_p[s] = theta_min;
				E_stab_p[1 * Dsens.size() + s] = theta_max;
			}
		} else if (nlhs >= 6) {
			plhs[5] = mxCreateNumericMatrix(gc.Solution().size(), 2, mxDOUBLE_CLASS, mxREAL);
			double* E_stab_p = (double*) mxGetPr(plhs[5]);

			std::vector<double> d(gc.Solution().size());
			double theta_min;
			double theta_max;

			for (unsigned int ei = 0; ei < d.size(); ++ei) {
				std::fill(d.begin(), d.end(), 0.0);
				d[ei] = 1.0;
				gc.ComputeStabilityRangeLP(d, theta_min, theta_max);
				E_stab_p[ei] = theta_min;
				E_stab_p[1 * d.size() + ei] = theta_max;
			}
		}
	}
}


